import test_typing_ext as t
import sys
import pytest
import platform

def test01_parameterize_generic():
    assert str(type(t.Wrapper[int]) == 't.Wrapper[int]')
    if platform.python_implementation() != 'PyPy':
        assert issubclass(t.WrapperFoo, t.Wrapper)
        assert t.WrapperFoo.__bases__ == (t.Wrapper,)
        assert t.WrapperFoo.__orig_bases__ == (t.Wrapper[t.Foo],)




def pytest_addoption(parser):
    parser.addoption('--enable-slow-tests',
                     action='store_true',
                     dest="enable-slow-tests",
                     default=False,
                     help="enable long-running tests")



#!/usr/bin/env python3

# With no command line flag, this prints the nanobind version.
# With flags -w semver, this writes the new version to where it's needed.

import os
import re
import sys

# Parse the header file <nanobind/nanobind.h> and print the version.
def get_version(root):
    major = ''
    minor = ''
    patch = ''
    dev = ''
    with open(os.path.join(root, "include/nanobind/nanobind.h"), 'r') as f:
        for line in f:
            if m := re.match(r'#define NB_VERSION_(.+)', line):
                if m_major := re.match(r'MAJOR\s+([0-9]+)', m.group(1)):
                    major = m_major.group(1)
                if m_minor := re.match(r'MINOR\s+([0-9]+)', m.group(1)):
                    minor = m_minor.group(1)
                if m_patch := re.match(r'PATCH\s+([0-9]+)', m.group(1)):
                    patch = m_patch.group(1)
                if m_dev := re.match(r'DEV\s+([0-9]+)', m.group(1)):
                    dev = m_dev.group(1)

    version_core = '.'.join([major, minor, patch])
    if int(dev) > 0:
        print(version_core, '-dev', dev, sep='')
    else:
        print(version_core)

# Write the semantic version to nanobind.h, pyproject.toml, __init__.py,
# and docs/bazel.rst.
# The semver string must be either 'X.Y.Z' or 'X.Y.Z-devN', where X, Y, Z are
# non-negative integers and N is a positive integer.
def write_version(root, semver):
    major = 0
    minor = 0
    patch = 0
    dev = 0
    try:
        beginning, middle, end = semver.split('.', maxsplit=2)
        major = int(beginning)
        minor = int(middle)
        if m := re.match(r'([0-9]+)-dev([1-9][0-9]*)', end):
            patch = int(m.group(1))
            dev = int(m.group(2))
        else:
            patch = int(end)
    except:
        print("Invalid version: '", semver, "'", sep='', file=sys.stderr)
        print("Valid examples: '1.2.3' or '1.2.3-dev4'", file=sys.stderr)
        return

    # Write to nanobind.h
    with open(os.path.join(root, "include/nanobind/nanobind.h"), "r+") as f:
        contents = f.read()
        contents = re.sub(r'#define NB_VERSION_MAJOR\s+[0-9]+',
                          r'#define NB_VERSION_MAJOR ' + str(major),
                          contents, count=1)
        contents = re.sub(r'#define NB_VERSION_MINOR\s+[0-9]+',
                          r'#define NB_VERSION_MINOR ' + str(minor),
                          contents, count=1)
        contents = re.sub(r'#define NB_VERSION_PATCH\s+[0-9]+',
                          r'#define NB_VERSION_PATCH ' + str(patch),
                          contents, count=1)
        contents = re.sub(r'#define NB_VERSION_DEV\s+[0-9]+',
                          r'#define NB_VERSION_DEV   ' + str(dev),
                          contents, count=1)
        f.seek(0)
        f.truncate()
        f.write(contents)

    # Write to pyproject.toml
    with open(os.path.join(root, "pyproject.toml"), "r+") as f:
        contents = f.read()
        contents = re.sub(r'version\s+=\s+"[^"]+"',
                          r'version = "' + semver + '"',
                          contents, count=1)
        f.seek(0)
        f.truncate()
        f.write(contents)

    # Write to __init__.py
    with open(os.path.join(root, "src/__init__.py"), "r+") as f:
        contents = f.read()
        contents = re.sub(r'__version__\s+=\s+"[^"]+"',
                          r'__version__ = "' + semver + '"',
                          contents, count=1)
        f.seek(0)
        f.truncate()
        f.write(contents)

    # write to docs/bazel.rst, but only if `semver` is not a dev release.
    # This is because documentation is scoped only to the latest stable release.
    if "dev" not in semver:
        with open(os.path.join(root, "docs/bazel.rst"), "r+") as f:
            contents = f.read()
            contents = re.sub(
                r"nanobind\s+v\d+(\.\d+)+",
                r"nanobind v" + semver,
                contents,
                count=1,
            )
            contents = re.sub(
                r'"nanobind_bazel", version = "\d+(\.\d+)+"',
                r'"nanobind_bazel", version = "' + semver + '"',
                contents,
                count=1,
            )
            f.seek(0)
            f.truncate()
            f.write(contents)


def main():
    root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    if len(sys.argv) == 1:
        get_version(root)
    elif len(sys.argv) == 3 and sys.argv[1] == '-w':
        write_version(root, sys.argv[2])
    else:
        print("Usage: ", sys.argv[0], file=sys.stderr)
        print("   or: ", sys.argv[0], "-w X.Y.Z", file=sys.stderr)
        print("   or: ", sys.argv[0], "-w X.Y.Z-devN", file=sys.stderr)


if __name__ == '__main__':
    main()



import pytest
import sys
import platform

import test_stl_bind_map_ext as t


def test_map_string_double(capfd):
    mm = t.MapStringDouble()
    mm["a"] = 1
    mm["b"] = 2.5

    assert list(mm) == ["a", "b"]
    assert "b" in mm
    assert 123 not in mm
    assert mm["b"] == 2.5

    assert "c" not in mm
    with pytest.raises(KeyError):
        mm["c"]
    assert "c" not in mm

    # Copy constructor
    mm2 = t.MapStringDouble(mm)
    assert list(mm2) == ["a", "b"]
    assert "b" in mm2
    assert "c" not in mm2
    assert 123 not in mm2
    assert mm2["b"] == 2.5
    assert mm == mm2
    assert not (mm != mm2)
    del mm2["b"]
    assert not (mm == mm2)
    assert mm != mm2
    assert len(mm2) == 1
    mm2.clear()
    assert len(mm2) == 0
    assert repr(mm) == "test_stl_bind_map_ext.MapStringDouble({'a': 1.0, 'b': 2.5})"

    with pytest.raises(TypeError):
        mm2.update({"a" : "b"})
    captured = capfd.readouterr().err.strip()
    ref = "nanobind: implicit conversion from type 'dict' to type 'test_stl_bind_map_ext.MapStringDouble' failed!"

    # Work around Pytest-related flakiness (https://github.com/pytest-dev/pytest/issues/10843)
    if platform.system() == 'Windows':
        assert captured == ref or captured == ''
    else:
        assert captured == ref

    mm2.update({"a" : 2.5})
    assert len(mm2) == 1

    # Construction from an iterable
    mm3 = t.MapStringDouble(
        { "a" : 1, "b" : 2.5 })
    assert list(mm3) == ["a", "b"]
    assert mm3["b"] == 2.5

    # Check that keys, values, items are views, not merely iterable
    keys = mm.keys()
    values = mm.values()
    items = mm.items()
    assert list(keys) == ["a", "b"]
    assert len(keys) == 2
    assert "a" in keys
    assert "c" not in keys
    assert 123 not in keys
    assert list(items) == [("a", 1), ("b", 2.5)]
    assert len(items) == 2
    assert ("b", 2.5) in items
    assert "hello" not in items
    assert ("b", 2.5, None) not in items
    assert list(values) == [1, 2.5]
    assert len(values) == 2
    assert 1 in values
    assert 2 not in values

    # Check that views update when the map is updated
    mm["c"] = -1
    assert list(keys) == ["a", "b", "c"]
    assert list(values) == [1, 2.5, -1]
    assert list(items) == [("a", 1), ("b", 2.5), ("c", -1)]

    um = t.UnorderedMapStringDouble()
    um["ua"] = 1.1
    um["ub"] = 2.6

    assert sorted(list(um)) == ["ua", "ub"]
    assert list(um.keys()) == list(um)
    assert sorted(list(um.items())) == [("ua", 1.1), ("ub", 2.6)]
    assert list(zip(um.keys(), um.values())) == list(um.items())

    assert type(keys).__qualname__ == 'MapStringDouble.KeyView'
    assert type(values).__qualname__ == 'MapStringDouble.ValueView'
    assert type(items).__qualname__ == 'MapStringDouble.ItemView'

    d = "dict"

    assert t.MapStringDouble.__init__.__doc__ == \
"""__init__(self) -> None
__init__(self, arg: test_stl_bind_map_ext.MapStringDouble) -> None
__init__(self, arg: %s[str, float], /) -> None

Overloaded function.

1. ``__init__(self) -> None``

Default constructor

2. ``__init__(self, arg: test_stl_bind_map_ext.MapStringDouble) -> None``

Copy constructor

3. ``__init__(self, arg: %s[str, float], /) -> None``

Construct from a dictionary""" % (d, d)


def test_map_string_double_const():
    mc = t.MapStringDoubleConst()
    mc["a"] = 10
    mc["b"] = 20.5

    umc = t.UnorderedMapStringDoubleConst()
    umc["a"] = 11
    umc["b"] = 21.5

    str(umc)


def test_maps_with_noncopyable_values():
    if not hasattr(t, 'get_mnc'):
        return

    # std::map
    mnc = t.get_mnc(5)
    for i in range(1, 6):
        assert mnc[i].value == 10 * i

    vsum = 0
    for k, v in mnc.items():
        assert v.value == 10 * k
        vsum += v.value

    assert vsum == 150

    # std::unordered_map
    mnc = t.get_umnc(5)
    for i in range(1, 6):
        assert mnc[i].value == 10 * i

    vsum = 0
    for k, v in mnc.items():
        assert v.value == 10 * k
        vsum += v.value

    assert vsum == 150

    # nested std::map<std::vector>
    nvnc = t.get_nvnc(5)
    for i in range(1, 6):
        for j in range(0, 5):
            assert nvnc[i][j].value == j + 1

    # Note: maps do not have .values()
    for _, v in nvnc.items():
        for i, j in enumerate(v, start=1):
            assert j.value == i

    # nested std::map<std::map>
    nmnc = t.get_nmnc(5)
    for i in range(1, 6):
        for j in range(10, 60, 10):
            assert nmnc[i][j].value == 10 * j

    vsum = 0
    for _, v_o in nmnc.items():
        for k_i, v_i in v_o.items():
            assert v_i.value == 10 * k_i
            vsum += v_i.value

    assert vsum == 7500

    # nested std::unordered_map<std::unordered_map>
    numnc = t.get_numnc(5)
    for i in range(1, 6):
        for j in range(10, 60, 10):
            assert numnc[i][j].value == 10 * j

    vsum = 0
    for _, v_o in numnc.items():
        for k_i, v_i in v_o.items():
            assert v_i.value == 10 * k_i
            vsum += v_i.value

    assert vsum == 7500


def test_map_delitem():
    mm = t.MapStringDouble()
    mm["a"] = 1
    mm["b"] = 2.5

    assert list(mm) == ["a", "b"]
    assert list(mm.items()) == [("a", 1), ("b", 2.5)]
    del mm["a"]
    assert list(mm) == ["b"]
    assert list(mm.items()) == [("b", 2.5)]

    um = t.UnorderedMapStringDouble()
    um["ua"] = 1.1
    um["ub"] = 2.6

    assert sorted(list(um)) == ["ua", "ub"]
    assert sorted(list(um.items())) == [("ua", 1.1), ("ub", 2.6)]
    del um["ua"]
    assert sorted(list(um)) == ["ub"]
    assert sorted(list(um.items())) == [("ub", 2.6)]



import sys
import sysconfig
import dis
import pytest

# Note: these tests verify that CPython's adaptive specializing interpreter can
# optimize various expressions involving nanobind types. They are expected to
# be somewhat fragile across Python versions as the bytecode and specialization
# opcodes may change.

# Skip tests on PyPy and free-threaded Python
skip_tests = sys.implementation.name == "pypy" or \
    sysconfig.get_config_var("Py_GIL_DISABLED")

import test_classes_ext as t
def disasm(func):
    """Extract specialized opcode names from a function"""
    instructions = list(dis.get_instructions(func, adaptive=True))
    return [(instr.opname, instr.argval) for instr in instructions]

def warmup(fn):
    # Call the function a few times to ensure that it is specialized
    for _ in range(8):
        fn()

def count_op(ops, expected):
    hits = 0
    for opname, _ in ops:
        if opname == expected:
            hits += 1
    return hits

@pytest.mark.skipif(
    sys.version_info < (3, 14) or skip_tests,
    reason="Static attribute specialization requires CPython 3.14+")
def test_static_attribute_specialization():
    s = t.Struct
    def fn():
        return s.static_test

    ops = disasm(fn)
    print(ops)
    op_base = count_op(ops, "LOAD_ATTR")
    op_opt = (
        count_op(ops, "LOAD_ATTR_ADAPTIVE") +
        count_op(ops, "LOAD_ATTR_CLASS"))
    assert op_base == 1 and op_opt == 0

    warmup(fn)
    ops = disasm(fn)
    print(ops)

    op_base = count_op(ops, "LOAD_ATTR")
    op_opt = (
        count_op(ops, "LOAD_ATTR_ADAPTIVE") +
        count_op(ops, "LOAD_ATTR_CLASS"))
    assert op_base == 0 and op_opt == 1

@pytest.mark.skipif(
    sys.version_info < (3, 11) or skip_tests,
    reason="Method call specialization requires CPython 3.14+")
def test_method_call_specialization():
    s = t.Struct()
    def fn():
        return s.value()

    ops = disasm(fn)
    op_base = (
        count_op(ops, "LOAD_METHOD") +
        count_op(ops, "LOAD_ATTR"))
    op_opt = (
        count_op(ops, "LOAD_ATTR_METHOD_NO_DICT") +
        count_op(ops, "CALL_ADAPTIVE"))
    print(ops)
    assert op_base == 1 and op_opt == 0

    warmup(fn)
    ops = disasm(fn)
    print(ops)
    op_base = (
        count_op(ops, "LOAD_METHOD") +
        count_op(ops, "LOAD_ATTR"))
    op_opt = (
        count_op(ops, "LOAD_ATTR_METHOD_NO_DICT") +
        count_op(ops, "CALL_ADAPTIVE"))
    assert op_base == 0 and op_opt == 1


@pytest.mark.skipif(sys.version_info < (3, 11) or skip_tests,
    reason="Immutability requires Python 3.11+")
def test_immutability():
    # Test nb_method immutability
    method = t.Struct.value
    method_type = type(method)
    assert method_type.__name__ == "nb_method"
    with pytest.raises(TypeError, match="immutable"):
        method_type.test_attr = 123

    # Test metaclass immutability
    metaclass = type(t.Struct)
    assert metaclass.__name__.startswith("nb_type")
    with pytest.raises(TypeError, match="immutable"):
        metaclass.test_attr = 123



from urllib.request import urlopen
import tarfile
import subprocess

funcs: "set[str]" = set()

files = [
    ('https://downloads.python.org/pypy/pypy3.9-v7.3.11-macos_arm64.tar.bz2', 'pypy3.9-v7.3.11-macos_arm64/bin/libpypy3.9-c.dylib'),
    ('https://downloads.python.org/pypy/pypy3.11-v7.3.20-macos_arm64.tar.bz2', 'pypy3.11-v7.3.20-macos_arm64/bin/libpypy3.11-c.dylib'),
]

for f in files:
    fs = urlopen(f[0])
    ft = tarfile.open(fileobj=fs, mode="r|bz2")
    success = False
    for member in ft: # move to the next file each loop
        if member.name == f[1]:
            ft.extract(member, path='tmp')
            success = True
    assert success

    out = subprocess.check_output(['nm', '-gjU', 'tmp/' + f[1]])
    for line in out.decode().split('\n'):
        if line.startswith('_Py') or line.startswith('__Py'):
            funcs.add(line)

with open("darwin-ld-pypy.sym", "w") as f:
    for func in sorted(list(funcs)):
        f.write(f'-U {func}\n')
